<?php

namespace Drupal\site_integrator\Controller;

/**
 * Site integrator main controller.
 */

use Drupal\Core\Controller\ControllerBase;
use Drupal\site_integrator\Entity\IntegratedSiteInterface;
use GuzzleHttp\Exception\BadResponseException;
use GuzzleHttp\Exception\GuzzleException;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Class SiteIntegratorController.
 *
 * Site integrator main controller class.
 */
class SiteIntegratorController extends ControllerBase {

  /**
   * Site integrator proxy.
   *
   * Useful documentation:
   *
   *   - https://api.drupal.org/api/drupal/vendor!symfony!http-foundation!Request.php/class/Request/9.3.x
   *   - https://api.drupal.org/api/drupal/vendor%21symfony%21routing%21Route.php/class/Route/9.3.x
   *   - https://api.drupal.org/api/drupal/vendor%21symfony%21http-foundation%21ParameterBag.php/class/ParameterBag/9.3.x
   */
  public function siteIntegratorPage() {
    \Drupal::service('page_cache_kill_switch')->trigger();
    $request = \Drupal::request();

    $path = $request->getPathInfo();
    $page = [
      '#type' => 'markup',
      '#markup' => $this->t(
        'Integrated page not processed (@path).',
        ['@path' => $path]
      ),
      '#cache' => ['max-age' => 0],
    ];
    $integrated_site = site_integrator_get_integrated_site($path);
    if (!empty($integrated_site)) {
      // Check if disabled.
      if (!$integrated_site->isEnabled()) {
        $redirect = $integrated_site->getRedirection('disabled');
        if (!empty($redirect)) {
          return new RedirectResponse($redirect);
        }
        throw new NotFoundHttpException('Site path disabled.');
      }
      // Check current user access permission.
      $roles = $integrated_site->getRoles();
      if (!empty($roles)) {
        // We have a role-restricted site integration.
        $user_roles = \Drupal::currentUser()->getRoles();
        if (!array_intersect($roles, $user_roles)) {
          $redirect = $integrated_site->getRedirection('forbidden');
          if (!empty($redirect)) {
            return new RedirectResponse($redirect);
          }
          throw new HttpException(401, 'Access denied');
        }
      }
      // Extract sub-path elements.
      $intgr_path = $integrated_site->getPath();
      $full_path = preg_replace('#^\Q' . $intgr_path . '\E#', '', $path);
      $base_url = $integrated_site->getBaseUrl();
      // Trim multiple slashes.
      $site_url = preg_replace('~(?<!:|^)//+~', '/', $base_url . $full_path);

      $mode = $integrated_site->getIntegrationMode();
      $frame_id = uniqid($integrated_site->id() . '_');
      if (IntegratedSiteInterface::INTEGRATION_MODE_EXTERNAL_IFRAME == $mode) {
        $query = $request->getQueryString();
        if (!empty($query)) {
          $site_url .= '?' . $query;
        }
        $page = [
          '#type' => 'theme',
          '#theme' => 'site_integrator_frame',
          '#frame_id' => $frame_id,
          '#site_src' => $site_url,
          '#site_width' => $integrated_site->get('width') ?? NULL,
          '#site_height' => $integrated_site->get('height') ?? NULL,
          '#cache' => ['max-age' => 0],
        ];
      }
      else {
        // Check if current Drupal URL does not end with '/'.
        if (empty($full_path)) {
          // Redirect to the correct URL with ending slash otherwise relative
          // URL in integrated site will not match the correct integrated URLs.
          $redirect_url = $request->getBasePath() . $request->getPathInfo() . '/';
          $query = $request->getQueryString();
          if (!empty($query)) {
            $redirect_url .= '?' . $query;
          }
          return new RedirectResponse($redirect_url);
        }
        $method = strtolower($request->getMethod());

        $page = [
          '#type' => 'markup',
          '#markup' => $this->t(
            'Integrated page not processed (@path).',
            ['@path' => $path]
          ),
          '#cache' => ['max-age' => 0],
        ];

        try {
          // phpcs:disable
          // Doc:
          // Symfony\Component\HttpFoundation\Response: https://github.com/symfony/symfony/blob/7.1/src/Symfony/Component/HttpFoundation/Response.php
          // Symfony\Component\HttpFoundation\Response->headers: https://github.com/symfony/symfony/blob/7.1/src/Symfony/Component/HttpFoundation/ResponseHeaderBag.php
          // \Drupal::service('http_client'):
          // - https://docs.guzzlephp.org/en/stable/
          // - https://github.com/guzzle/guzzle/blob/7.9/src/Client.php
          // \Drupal::service('http_client')->request(): https://github.com/guzzle/psr7/blob/2.7/src/Response.php
          // phpcs:enable
          $http_client = \Drupal::service('http_client');
          $client_response = $http_client->request(
            $method,
            $site_url,
            [
              'query' => $request->query->all(),
              'headers' => $request->headers->all(),
              'verify' => $integrated_site->getVerify(),
            ]
          );
          $status_code = $client_response->getStatusCode();
        }
        catch (GuzzleException $exception) {
          if ($exception instanceof BadResponseException) {
            $client_response = $exception->getResponse();
            $status_code = $client_response->getStatusCode();
          }
          else {
            // Use a fake status code to signal an error (eg. timeout).
            $status_code = 1000;
          }
        }

        // See what to do with the client response.
        if ((200 <= $status_code) && (300 > $status_code)) {
          $passthrough = $integrated_site->getPassthroughList();
          // Only rework HTML content.
          if ((0 === strpos($client_response->getHeader('content-type')[0] ?? '', 'text/html'))
              && (!in_array($full_path, $passthrough))
          ) {
            // Rework the HTML, get site content.
            $body = $client_response->getBody() . '';
            if (IntegratedSiteInterface::INTEGRATION_MODE_INTERNAL_IFRAME == $mode) {
              $page = [
                '#type' => 'theme',
                '#theme' => 'site_integrator_frame',
                '#frame_id' => $frame_id,
                '#site_width' => $integrated_site->get('width') ?? NULL,
                '#site_height' => $integrated_site->get('height') ?? NULL,
                '#cache' => ['max-age' => 0],
                '#attached' => [
                  'drupalSettings' => [
                    'site_integrator_sites' => [
                      $frame_id => [
                        'frame_id ' => $frame_id,
                        'site_content' => str_replace(
                          ['&', '<', '>', '"', "\r", "\n"],
                          ['&amp;', '&lt;', '&gt;', '&quot;', '&#13;', '&#10;'],
                          $body
                        ),
                        'filters' => $integrated_site->getFilters(),
                        'site_width' => $integrated_site->get('width') ?? NULL,
                        'site_height' => $integrated_site->get('height') ?? NULL,
                      ],
                    ],
                  ],
                ],
              ];
            }
            elseif (IntegratedSiteInterface::INTEGRATION_MODE_MERGE == $mode) {
              // Rewrite HTML.
              // Remove unsupported tags.
              $body = preg_replace(
                '#</?(?:!DOCTYPE|html|head|meta|title|body)[^>]*>#si',
                '',
                $body
              );
              $page = [
                '#type' => 'markup',
                '#children' => $body,
                '#frame_id' => $frame_id,
                '#site_width' => $integrated_site->get('width') ?? NULL,
                '#site_height' => $integrated_site->get('height') ?? NULL,
                '#cache' => ['max-age' => 0],
                '#attached' => [
                  'drupalSettings' => [
                    'site_integrator_sites' => [
                      $frame_id => [
                        'frame_id ' => $frame_id,
                        'site_content' => str_replace(
                          ['&', '<', '>', '"', "\r", "\n"],
                          ['&amp;', '&lt;', '&gt;', '&quot;', '&#13;', '&#10;'],
                          $body
                        ),
                        'filters' => $integrated_site->getFilters(),
                        'site_width' => $integrated_site->get('width') ?? NULL,
                        'site_height' => $integrated_site->get('height') ?? NULL,
                      ],
                    ],
                  ],
                ],
              ];
            }
          }
          else {
            // Not HTML, pass through.
            $response = new Response();
            $response->setContent($client_response->getBody());
            $response->setStatusCode($status_code);
            $headers = $client_response->getHeaders();
            foreach ($headers as $header => $header_value) {
              $response->headers->set($header, $header_value);
            }
            return $response;
          }
        }
        else {
          // Log errors.
          $message =
            'Client status code: '
            . (1000 == $status_code ? 'Guzzle exception' : $status_code)
            . "\nURL: "
            . $site_url
            . "\n"
            . (!empty($client_response) ? $client_response->getBody() : (!empty($exception) ? $exception->getMessage() : ''));
          \Drupal::logger('site_integrator')->warning($message);
          // Manage redirections.
          $redirect = $integrated_site->getRedirection($status_code);
          if (empty($redirect)) {
            $redirect = $integrated_site->getRedirection(
              ((int) ($status_code / 100)) . 'xx'
            );
          }
          if (!empty($redirect)) {
            return new RedirectResponse($redirect);
          }
          // Generate a blank page.
          $response = new Response();
          $response->setStatusCode(Response::HTTP_OK);
          $response->setContent('');
          return $response;
        }
      }
    }

    return $page;
  }

}
